@@ -2,7 +2,7 @@
 '''
 test_dugong.py - Unit tests for dugong.py - run with py.test
 
-Copyright (c) Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2014 Nikolaus Rath <Nikolaus.org>
 
 This module may be distributed under the terms of the Python Software Foundation
 License Version 2.  The complete license text may be retrieved from
@@ -24,7 +24,7 @@ if __name__ == '__main__':
 from dugong import (HTTPConnection, BodyFollowing, CaseInsensitiveDict, _join,
                     ConnectionClosed)
 import dugong
-from http.server import BaseHTTPRequestHandler, _quote_html
+from http.server import BaseHTTPRequestHandler
 from io import TextIOWrapper
 from base64 import b64encode
 import http.client
@@ -34,6 +34,7 @@ import time
 import ssl
 import re
 import os
+import html
 import hashlib
 import threading
 import socketserver
@@ -466,6 +467,123 @@ def test_read_identity(conn):
     assert conn.readall() == DUMMY_DATA[:512]
     assert not conn.response_pending()
 
+def test_conn_close_1(conn, monkeypatch):
+    # Regular read
+    data_size = 500
+    conn._rbuf = dugong._Buffer(int(4/5*data_size))
+
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Connection", 'close')
+        self.end_headers()
+        self.wfile.write(DUMMY_DATA[:data_size])
+        self.rfile.close()
+        self.wfile.close()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/whatever')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert conn.readall() == DUMMY_DATA[:data_size]
+
+    with pytest.raises(ConnectionClosed):
+        conn.send_request('GET', '/whatever')
+        conn.read_response()
+
+def test_conn_close_2(conn, monkeypatch):
+    # Readinto
+    data_size = 500
+    conn._rbuf = dugong._Buffer(int(4/5*data_size))
+
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Connection", 'close')
+        self.end_headers()
+        self.wfile.write(DUMMY_DATA[:data_size])
+        self.rfile.close()
+        self.wfile.close()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/whatever')
+    resp = conn.read_response()
+    assert resp.status == 200
+    buf = memoryview(bytearray(2*data_size))
+    got = conn.readinto(buf)
+    got += conn.readinto(buf[got:])
+    assert got == data_size
+    assert buf[:data_size] == DUMMY_DATA[:data_size]
+    assert conn.readinto(buf[got:]) == 0
+
+    with pytest.raises(ConnectionClosed):
+        conn.send_request('GET', '/whatever')
+        conn.read_response()
+
+def test_conn_close_3(conn, monkeypatch):
+    # Server keeps reading
+    data_size = 500
+    conn._rbuf = dugong._Buffer(int(4/5*data_size))
+
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Connection", 'close')
+        self.end_headers()
+        self.wfile.write(DUMMY_DATA[:data_size])
+        self.wfile.close()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/whatever')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert conn.readall() == DUMMY_DATA[:data_size]
+
+    conn.send_request('GET', '/whatever')
+    assert_raises(ConnectionClosed, conn.read_response)
+
+def test_conn_close_4(conn, monkeypatch):
+    # Content-Length should take precedence
+    data_size = 500
+    conn._rbuf = dugong._Buffer(int(4/5*data_size))
+
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Content-Length", str(data_size))
+        self.send_header("Connection", 'close')
+        self.end_headers()
+        self.wfile.write(DUMMY_DATA[:data_size+10])
+        self.wfile.close()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/whatever')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert conn.readall() == DUMMY_DATA[:data_size]
+
+def test_conn_close_5(conn, monkeypatch):
+    # Pipelining
+    data_size = 500
+    conn._rbuf = dugong._Buffer(int(4/5*data_size))
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Content-Length", str(data_size))
+        self.send_header("Connection", 'close')
+        self.end_headers()
+        self.wfile.write(DUMMY_DATA[:data_size])
+        self.wfile.close()
+        self.rfile.close()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/whatever_one')
+    conn.send_request('GET', '/whatever_two')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert conn.readall() == DUMMY_DATA[:data_size]
+    assert_raises(dugong.ConnectionClosed, conn.read_response)
+
 @pytest.mark.no_ssl
 def test_exhaust_buffer(conn):
     conn._rbuf = dugong._Buffer(600)
@@ -1046,9 +1164,12 @@ class MockRequestHandler(BaseHTTPRequestHandler):
             message = shortmsg
         explain = longmsg
         self.log_error("code %d, message %s", code, message)
-        # using _quote_html to prevent Cross Site Scripting attacks (see bug #1100201)
-        content = (self.error_message_format % {'code': code, 'message': _quote_html(message),
-                                               'explain': explain}).encode('utf-8', 'replace')
+        # HTML encode to prevent Cross Site Scripting attacks (see bug #1100201)
+        content = (self.error_message_format % {
+            'code': code,
+            'message': html.escape(message, quote=False),
+            'explain': explain
+        }).encode('utf-8', 'replace')
         self.send_response(code, message)
         self.send_header("Content-Type", self.error_content_type)
         self.send_header("Content-Length", str(len(content)))
